import hashlib
import json
import re
import ssl
import subprocess
import time
from pathlib import Path
from typing import Optional

from daktari.check import Check, CheckResult
from daktari.command_utils import CommandErrorException, CommandNotFoundException, run_command
from daktari.os import OS


def command_failure_summary(error) -> str:
    if isinstance(error, CommandErrorException):
        output = f"{error.stderr}\n{error.stdout}".strip()
        for line in output.splitlines():
            cleaned_line = line.strip()
            if cleaned_line:
                return cleaned_line
    return str(error)


def command_output(result) -> str:
    return f"{result.stderr}\n{result.stdout}".strip()


def get_available_ios_simulators():
    payload = json.loads(run_command(["xcrun", "simctl", "list", "devices", "available", "--json"]).stdout)
    devices = []

    for runtime, runtime_devices in payload.get("devices", {}).items():
        for raw_device in runtime_devices:
            name = str(raw_device.get("name", ""))
            runtime_identifier = str(runtime)
            is_ios_runtime = runtime_identifier.startswith("com.apple.CoreSimulator.SimRuntime.iOS-")
            if not raw_device.get("isAvailable", True) or not is_ios_runtime:
                continue

            devices.append(
                {
                    "udid": str(raw_device.get("udid", "")),
                    "name": name,
                    "runtime": runtime,
                    "booted": str(raw_device.get("state", "")) == "Booted",
                    "dataPath": str(raw_device.get("dataPath", "")),
                }
            )

    if len(devices) == 0:
        raise CommandErrorException("No iOS simulators found", 1, "", "No available iOS simulators found")

    return sorted(devices, key=lambda device: (device["runtime"], device["name"]), reverse=True)


def get_certificate_sha256_from_pem(pem_path) -> str:
    pem_contents = Path(pem_path).read_text()
    cert_der = ssl.PEM_cert_to_DER_cert(pem_contents)
    return hashlib.sha256(cert_der).hexdigest()


def get_available_android_avds():
    avds = [line.strip() for line in run_command(["emulator", "-list-avds"]).stdout.splitlines() if line.strip()]
    if len(avds) == 0:
        raise CommandErrorException("No Android AVDs found", 1, "", "No Android AVDs found")

    return sorted(avds)


def get_booted_android_serial_for_avd(avd_name):
    devices = run_command(["adb", "devices"]).stdout.splitlines()[1:]
    for line in devices:
        parts = line.split()
        if len(parts) < 2:
            continue

        serial = parts[0]
        status = parts[1]
        if not serial.startswith("emulator-") or status != "device":
            continue

        try:
            emu_output = run_command(["adb", "-s", serial, "emu", "avd", "name"]).stdout
        except (CommandErrorException, CommandNotFoundException):
            continue

        emu_output_lines = [line.strip() for line in emu_output.replace("\r", "").splitlines() if line.strip()]
        reported_avd_name = next((line for line in emu_output_lines if line != "OK"), "")
        if reported_avd_name == avd_name:
            return serial

    return None


def wait_for_android_device_boot(serial, timeout_seconds=240):
    deadline = time.time() + timeout_seconds
    while True:
        if time.time() > deadline:
            raise CommandErrorException(
                "Timed out waiting for Android emulator to boot",
                1,
                "",
                f"Timed out waiting for Android emulator {serial} to boot",
            )

        try:
            result = run_command(["adb", "-s", serial, "shell", "getprop", "sys.boot_completed"])
            if result.stdout.strip().replace("\r", "") == "1":
                return
        except (CommandErrorException, CommandNotFoundException):
            pass

        time.sleep(2)


def get_android_system_property(serial, property_name) -> str:
    try:
        result = run_command(["adb", "-s", serial, "shell", "getprop", property_name])
        return result.stdout.strip().replace("\r", "")
    except (CommandErrorException, CommandNotFoundException):
        return ""


def get_android_bootstrap_skip_reason(serial, avd_name, adb_root_output) -> Optional[str]:
    build_type = get_android_system_property(serial, "ro.build.type")
    build_tags = get_android_system_property(serial, "ro.build.tags")
    debuggable = get_android_system_property(serial, "ro.debuggable")
    lower_root_output = adb_root_output.lower()

    if "adbd cannot run as root in production builds" in lower_root_output:
        return (
            f"{avd_name} is running a non-rootable production image "
            f"(ro.build.type={build_type or 'unknown'}, ro.build.tags={build_tags or 'unknown'}, "
            f"ro.debuggable={debuggable or 'unknown'}). Use a Google APIs or AOSP userdebug image instead."
        )

    if debuggable == "0" or build_type == "user" or build_tags == "release-keys":
        return (
            f"{avd_name} is running a non-rootable production image "
            f"(ro.build.type={build_type or 'unknown'}, ro.build.tags={build_tags or 'unknown'}, "
            f"ro.debuggable={debuggable or 'unknown'}). Use a Google APIs or AOSP userdebug image instead."
        )

    return None


def start_android_avd_for_check(avd_name, emulator_log_path="/tmp/daktari-mobile-emulator.log"):
    with Path(emulator_log_path).open("a", encoding="utf-8") as log_file:
        subprocess.Popen(["emulator", "-avd", avd_name, "-writable-system"], stdout=log_file, stderr=subprocess.STDOUT)

    deadline = time.time() + 240
    while True:
        serial = get_booted_android_serial_for_avd(avd_name)
        if serial is not None:
            wait_for_android_device_boot(serial)
            return serial

        if time.time() > deadline:
            raise CommandErrorException(
                "Timed out waiting for Android emulator to appear",
                1,
                "",
                f"Timed out waiting for emulator '{avd_name}' to appear",
            )

        time.sleep(2)


def android_hosts_mapping_present(serial, host_alias) -> bool:
    host_alias_regex = re.escape(host_alias)
    grep_pattern = f"^10\\.0\\.2\\.2[[:space:]]+{host_alias_regex}($|[[:space:]])"
    try:
        run_command(
            [
                "adb",
                "-s",
                serial,
                "shell",
                f"grep -Eq '{grep_pattern}' /system/etc/hosts",
            ]
        )
        return True
    except (CommandErrorException, CommandNotFoundException):
        return False


class MobileIosPrerequisitesReady(Check):
    name = "mobile.ios.prerequisites"
    suggestions = {OS.OS_X: """
            Common fixes:
            - Install Xcode and launch it at least once to finish setup
            - Install an iOS simulator runtime in Xcode Settings > Platforms
            - Install mkcert and run <cmd>mkcert -install</cmd>
            """}

    def check(self) -> CheckResult:
        try:
            run_command(["mkcert", "-CAROOT"])
            run_command(["xcrun", "simctl", "list", "devices", "available", "--json"])
            get_available_ios_simulators()
            return self.passed("iOS mobile bootstrap prerequisites are met")
        except (CommandErrorException, CommandNotFoundException) as error:
            failure_reason = command_failure_summary(error)
            return self.failed(f"iOS mobile bootstrap prerequisites are <not/> met: {failure_reason}")


class MobileAndroidPrerequisitesReady(Check):
    name = "mobile.android.prerequisites"
    suggestions = {OS.GENERIC: """
            Common fixes:
            - In Android Studio, open SDK Manager > SDK Tools and install Android SDK Platform-Tools (adb)
              and Android Emulator
            - Create at least one Android Virtual Device (AVD)
            - Add Android SDK paths to your shell profile PATH (for example
              .../Android/sdk/platform-tools and .../Android/sdk/emulator)
            - Install mkcert and run <cmd>mkcert -install</cmd>
            """}

    def check(self) -> CheckResult:
        try:
            run_command(["mkcert", "-CAROOT"])
            run_command(["adb", "devices"])
            get_available_android_avds()
            return self.passed("Android mobile bootstrap prerequisites are met")
        except (CommandErrorException, CommandNotFoundException) as error:
            failure_reason = command_failure_summary(error)
            return self.failed(f"Android mobile bootstrap prerequisites are <not/> met: {failure_reason}")


class MobileIosBootstrapReady(Check):
    name = "mobile.ios.bootstrapReady"
    depends_on = [MobileIosPrerequisitesReady]

    def __init__(self, bootstrap_command: str = "task mobile:bootstrap:ios"):
        self.suggestions = {OS.OS_X: f"""
            Complete iOS simulator setup:
            <cmd>{bootstrap_command}</cmd>
            """}

    def check(self) -> CheckResult:
        try:
            simulators = get_available_ios_simulators()

            root_ca_path = Path(run_command(["mkcert", "-CAROOT"]).stdout.strip()) / "rootCA.pem"
            if not root_ca_path.exists():
                return self.failed("iOS mobile bootstrap setup is not complete")

            root_ca_sha256 = get_certificate_sha256_from_pem(root_ca_path)
            missing_simulators = []

            for simulator in simulators:
                trust_store_path = (
                    Path(simulator["dataPath"]) / "private/var/protected/trustd/private/TrustStore.sqlite3"
                )
                if not trust_store_path.exists():
                    missing_simulators.append(simulator["name"])
                    continue

                trust_query = f"select count(*) from tsettings where lower(hex(sha256)) = '{root_ca_sha256}';"
                trust_query_result = run_command(["sqlite3", str(trust_store_path), trust_query]).stdout.strip()
                if not trust_query_result.isdigit() or int(trust_query_result) < 1:
                    missing_simulators.append(simulator["name"])

            if len(missing_simulators) > 0:
                missing_names = ", ".join(sorted(set(missing_simulators)))
                return self.failed(f"iOS mobile bootstrap setup is not complete: missing on {missing_names}")

            return self.passed("iOS mobile bootstrap setup is complete")
        except (CommandErrorException, CommandNotFoundException):
            return self.failed("iOS mobile bootstrap setup is <not/> complete")


class MobileAndroidBootstrapReady(Check):
    name = "mobile.android.bootstrapReady"
    depends_on = [MobileAndroidPrerequisitesReady]

    def __init__(
        self,
        local_certificate_path,
        android_certificate_path,
        host_alias,
        bootstrap_command: str = "task mobile:bootstrap:android",
        emulator_log_path: str = "/tmp/daktari-mobile-emulator.log",
    ):
        self.local_certificate_path = local_certificate_path
        self.android_certificate_path = android_certificate_path
        self.host_alias = host_alias
        self.emulator_log_path = emulator_log_path
        self.suggestions = {OS.GENERIC: f"""
            Complete Android emulator setup for all available AVDs:
            <cmd>{bootstrap_command}</cmd>
            """}

    def check(self) -> CheckResult:
        try:
            avds = get_available_android_avds()

            local_cert_path = Path(self.local_certificate_path)
            android_cert_path = Path(self.android_certificate_path)
            if not local_cert_path.exists() or not android_cert_path.exists():
                return self.failed("Android mobile bootstrap setup is not complete")

            if local_cert_path.read_bytes() != android_cert_path.read_bytes():
                return self.failed("Android mobile bootstrap setup is not complete")

            missing_avds = []
            unsupported_avds = []
            for avd_name in avds:
                serial = get_booted_android_serial_for_avd(avd_name)
                started_by_check = serial is None

                if serial is None:
                    serial = start_android_avd_for_check(avd_name, self.emulator_log_path)
                else:
                    wait_for_android_device_boot(serial)

                try:
                    adb_root_result = run_command(["adb", "-s", serial, "root"])
                    wait_for_android_device_boot(serial)

                    skip_reason = get_android_bootstrap_skip_reason(serial, avd_name, command_output(adb_root_result))
                    if skip_reason is not None:
                        unsupported_avds.append(avd_name)
                        continue

                    hosts_ok = android_hosts_mapping_present(serial, self.host_alias)
                    if not hosts_ok:
                        missing_avds.append(avd_name)
                finally:
                    if started_by_check:
                        try:
                            run_command(["adb", "-s", serial, "emu", "kill"])
                        except (CommandErrorException, CommandNotFoundException):
                            pass

            if len(unsupported_avds) == len(avds) and len(avds) > 0:
                return self.failed(
                    "Android mobile bootstrap setup is not complete: all available AVDs are using non-rootable "
                    "production images; create a Google APIs or AOSP userdebug AVD"
                )

            if len(missing_avds) > 0:
                missing_names = ", ".join(missing_avds)
                return self.failed(f"Android mobile bootstrap setup is not complete: missing on {missing_names}")

            return self.passed("Android mobile bootstrap setup is complete")
        except (CommandErrorException, CommandNotFoundException):
            return self.failed("Android mobile bootstrap setup is <not/> complete")



from daktari.check import Check, CheckResult
from daktari.os import OS


class PnpmInstalled(Check):
    name = "pnpm.installed"

    suggestions = {
        OS.OS_X: "<cmd>brew install pnpm</cmd>",
        OS.GENERIC: "<cmd>npm install -g pnpm</cmd>",
    }

    def check(self) -> CheckResult:
        return self.verify_install("pnpm")



import logging
import os
from dataclasses import dataclass
from typing import List, Optional
from urllib.parse import urlparse

from daktari.check import Check, CheckResult
from daktari.command_utils import can_run_command
from daktari.file_utils import file_exists
from daktari.os import OS

PLACEHOLDER_TOKEN = "UPDATE_WITH_TOKEN"


@dataclass
class NpmrcScope:
    name: str
    registry: str
    require_auth_token: bool = False


def get_npmrc_path() -> str:
    return os.path.expanduser("~/.npmrc")


def get_npmrc_contents() -> List[str]:
    npmrc_path = get_npmrc_path()
    if not file_exists(npmrc_path):
        raise Exception(f"{npmrc_path} does not exist")

    try:
        with open(npmrc_path, "r") as npmrc_file:
            return npmrc_file.readlines()
    except Exception:
        logging.error(f"Exception reading {npmrc_path}", exc_info=True)
        raise Exception("Failed to read npmrc")


def get_registry_host(registry_url: str) -> str:
    parsed = urlparse(registry_url)
    return parsed.netloc


def npmrc_contains_scope_registry(lines: List[str], scope: NpmrcScope) -> bool:
    expected_line = f"@{scope.name}:registry={scope.registry}"
    for line in lines:
        if line.strip() == expected_line:
            return True
    return False


def npmrc_contains_auth_token(lines: List[str], registry_url: str) -> bool:
    host = get_registry_host(registry_url)
    auth_token_prefix = f"//{host}/:_authToken="
    for line in lines:
        stripped = line.strip()
        if stripped.startswith(auth_token_prefix):
            token_value = stripped[len(auth_token_prefix) :]
            if token_value and token_value != PLACEHOLDER_TOKEN:
                return True
    return False


def npmrc_scope_is_configured(lines: List[str], scope: NpmrcScope) -> bool:
    if not npmrc_contains_scope_registry(lines, scope):
        return False
    if scope.require_auth_token and not npmrc_contains_auth_token(lines, scope.registry):
        return False
    return True


def get_npmrc_suggestion(scope: NpmrcScope) -> str:
    lines = [f"@{scope.name}:registry={scope.registry}"]
    if scope.require_auth_token:
        host = get_registry_host(scope.registry)
        lines.append(f"//{host}/:_authToken={PLACEHOLDER_TOKEN}")
    return "\n".join(lines)


class NpmrcScopeConfigured(Check):
    name = "npmrc.scopeConfigured"

    def __init__(self, scope: NpmrcScope, token_instructions: Optional[str] = None):
        self.scope = scope
        self.npmrc_suggestion = get_npmrc_suggestion(scope)
        token_instruction_string = f"\n\n{token_instructions}" if token_instructions else ""
        self.suggestions = {OS.GENERIC: f"""Add the lines below to ~/.npmrc:

{self.npmrc_suggestion}{token_instruction_string}"""}

    def check(self) -> CheckResult:
        try:
            lines = get_npmrc_contents()
        except Exception as e:
            return self.failed(str(e))

        if not npmrc_scope_is_configured(lines, self.scope):
            return self.failed(f"Scope {self.scope.name} not configured in npmrc")

        return self.passed(f"Scope {self.scope.name} configured in npmrc")


class NpmrcGithubTokenValid(Check):
    name = "npmrc.githubTokenValid"
    depends_on = [NpmrcScopeConfigured]

    def __init__(self, scope_name: str, test_package: str):
        self.scope_name = scope_name
        self.test_package = test_package
        self.suggestions = {
            OS.GENERIC: "Please check the token was copied correctly from GitHub."
            " Ensure the token hasn't expired, or has been revoked."
            " Also, ensure it has the correct permissions to read packages."
        }

    def check(self) -> CheckResult:
        package_spec = f"@{self.scope_name}/{self.test_package}"
        command = f"npm view {package_spec} version"
        logging.debug(f"Checking npm registry access with: {command}")

        if can_run_command(command):
            return self.passed(f"GitHub npm token is valid (can access {package_spec})")
        else:
            return self.failed(f"GitHub npm token is not valid (cannot access {package_spec})")



import os
from typing import List, Set, Type, Union

from daktari.check import Check
from daktari.collection_utils import flatten


class CyclicCheckException(Exception):
    def __init__(self, message):
        super().__init__(message)


def check_for_cycles(check: Union[Check, Type[Check]], prev_parents: Set[str]):
    if check.name in prev_parents:
        raise CyclicCheckException(f"Check [{check.name}] has cyclic dependencies")
    for sub_check in check.depends_on:
        check_for_cycles(sub_check, prev_parents.union({check.name}))


def get_all_dependent_check_names(check: Union[Check, Type[Check]]) -> Set[str]:
    check_for_cycles(check, set())
    return _get_all_dependent_check_names_recursive(check)


def _get_all_dependent_check_names_recursive(check: Union[Check, Type[Check]]) -> Set[str]:
    sub_dependents = [_get_all_dependent_check_names_recursive(dep) for dep in check.depends_on]
    flat_sub_dependents = flatten(sub_dependents)
    return flat_sub_dependents.union({dep.name for dep in check.depends_on})


def filter_checks_by_name(
    checks: List[Check],
    requested_check_names: Set[str],
    unknown_check_error_prefix: str = "Unknown Daktari check names",
) -> List[Check]:
    available_checks_by_name = {check.name: check for check in checks}

    unknown_check_names = sorted(requested_check_names - set(available_checks_by_name.keys()))
    if len(unknown_check_names) > 0:
        raise ValueError(f"{unknown_check_error_prefix}: {', '.join(unknown_check_names)}")

    required_check_names = set()
    for requested_check_name in requested_check_names:
        requested_check = available_checks_by_name[requested_check_name]
        required_check_names.add(requested_check_name)
        required_check_names.update(get_all_dependent_check_names(requested_check))

    return [check for check in checks if check.name in required_check_names]


def filter_out_checks_by_name(checks: List[Check], ignored_check_names: Set[str]) -> List[Check]:
    checks_to_ignore = [
        check
        for check in checks
        if check.name in ignored_check_names
        or len(get_all_dependent_check_names(check).intersection(ignored_check_names)) > 0
    ]
    return [check for check in checks if check not in checks_to_ignore]


def filter_checks_by_env_var(checks: List[Check], env_var_name: str = "DAKTARI_ONLY_CHECKS") -> List[Check]:
    only_checks_csv = os.getenv(env_var_name, "").strip()
    if only_checks_csv == "":
        return checks

    requested_check_names = {check_name.strip() for check_name in only_checks_csv.split(",") if check_name.strip()}
    return filter_checks_by_name(
        checks,
        requested_check_names,
        unknown_check_error_prefix=f"Unknown Daktari check names in {env_var_name}",
    )



import json
import tempfile
import unittest
from unittest import mock

from daktari.check import CheckStatus
from daktari.checks.mobile import (
    MobileAndroidBootstrapReady,
    command_failure_summary,
    get_available_android_avds,
    get_available_ios_simulators,
)
from daktari.command_utils import CommandErrorException, SuccessfulCommandResult


class TestCommandFailureSummary(unittest.TestCase):
    def test_extracts_first_non_empty_line_from_command_error(self):
        error = CommandErrorException("failed", 1, "", "\nfirst line\nsecond line")

        self.assertEqual("first line", command_failure_summary(error))

    def test_falls_back_to_error_string(self):
        self.assertEqual("boom", command_failure_summary(Exception("boom")))


class TestMobileHelpers(unittest.TestCase):
    @mock.patch("daktari.checks.mobile.run_command")
    def test_get_available_ios_simulators_filters_non_ios_and_unavailable(self, mock_run_command):
        payload = {
            "devices": {
                "com.apple.CoreSimulator.SimRuntime.iOS-18-0": [
                    {
                        "name": "iPhone 16",
                        "udid": "ios-1",
                        "state": "Shutdown",
                        "isAvailable": True,
                        "dataPath": "/tmp/ios-1",
                    },
                    {
                        "name": "Unavailable",
                        "udid": "ios-2",
                        "state": "Shutdown",
                        "isAvailable": False,
                        "dataPath": "/tmp/ios-2",
                    },
                ],
                "com.apple.CoreSimulator.SimRuntime.tvOS-18-0": [
                    {
                        "name": "Apple TV",
                        "udid": "tvos-1",
                        "state": "Shutdown",
                        "isAvailable": True,
                        "dataPath": "/tmp/tvos-1",
                    }
                ],
            }
        }
        mock_run_command.return_value.stdout = json.dumps(payload)

        simulators = get_available_ios_simulators()

        self.assertEqual(1, len(simulators))
        self.assertEqual("iPhone 16", simulators[0]["name"])

    @mock.patch("daktari.checks.mobile.run_command")
    def test_get_available_android_avds_returns_sorted_list(self, mock_run_command):
        mock_run_command.return_value.stdout = "Pixel_8\n\nPixel_7\n"

        self.assertEqual(["Pixel_7", "Pixel_8"], get_available_android_avds())

    @mock.patch("daktari.checks.mobile.run_command")
    def test_get_available_android_avds_raises_when_none_available(self, mock_run_command):
        mock_run_command.return_value.stdout = "\n"

        with self.assertRaisesRegex(CommandErrorException, "No Android AVDs found"):
            get_available_android_avds()


class TestMobileAndroidBootstrapReady(unittest.TestCase):
    def test_fails_when_all_available_avds_are_non_rootable(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            local_certificate_path = f"{temp_dir}/local.pem"
            android_certificate_path = f"{temp_dir}/android.pem"
            with open(local_certificate_path, "wb") as local_cert_file:
                local_cert_file.write(b"same")
            with open(android_certificate_path, "wb") as android_cert_file:
                android_cert_file.write(b"same")

            check = MobileAndroidBootstrapReady(
                local_certificate_path,
                android_certificate_path,
                "local.example.test",
            )

            with (
                mock.patch(
                    "daktari.checks.mobile.get_available_android_avds",
                    return_value=["Medium_Phone_API_35"],
                ),
                mock.patch(
                    "daktari.checks.mobile.get_booted_android_serial_for_avd",
                    return_value="emulator-5554",
                ),
                mock.patch(
                    "daktari.checks.mobile.run_command",
                    return_value=SuccessfulCommandResult(
                        stdout="adbd cannot run as root in production builds\n",
                        stderr="",
                    ),
                ),
                mock.patch("daktari.checks.mobile.wait_for_android_device_boot"),
                mock.patch(
                    "daktari.checks.mobile.get_android_bootstrap_skip_reason",
                    return_value="Medium_Phone_API_35 is running a non-rootable production image",
                ),
            ):
                result = check.check()

        self.assertEqual(CheckStatus.FAIL, result.status)
        self.assertIn("all available AVDs are using non-rootable production images", result.summary)

    def test_ignores_non_rootable_avds_when_supported_avds_are_bootstrapped(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            local_certificate_path = f"{temp_dir}/local.pem"
            android_certificate_path = f"{temp_dir}/android.pem"
            with open(local_certificate_path, "wb") as local_cert_file:
                local_cert_file.write(b"same")
            with open(android_certificate_path, "wb") as android_cert_file:
                android_cert_file.write(b"same")

            check = MobileAndroidBootstrapReady(
                local_certificate_path,
                android_certificate_path,
                "local.example.test",
            )

            with (
                mock.patch(
                    "daktari.checks.mobile.get_available_android_avds",
                    return_value=["Pixel_8_API_35", "Medium_Phone_API_35"],
                ),
                mock.patch(
                    "daktari.checks.mobile.get_booted_android_serial_for_avd",
                    return_value="emulator-5554",
                ) as mock_get_serial,
                mock.patch(
                    "daktari.checks.mobile.run_command",
                    return_value=SuccessfulCommandResult(stdout="restarting adbd as root\n", stderr=""),
                ),
                mock.patch("daktari.checks.mobile.wait_for_android_device_boot"),
                mock.patch(
                    "daktari.checks.mobile.get_android_bootstrap_skip_reason",
                    side_effect=[None, "Medium_Phone_API_35 is running a non-rootable production image"],
                ),
                mock.patch(
                    "daktari.checks.mobile.android_hosts_mapping_present",
                    return_value=True,
                ),
            ):
                result = check.check()

        self.assertEqual(CheckStatus.PASS, result.status)
        self.assertEqual(2, mock_get_serial.call_count)


if __name__ == "__main__":
    unittest.main()



import unittest
from unittest import mock

from daktari.check import CheckStatus
from daktari.checks.npmrc import (
    NpmrcScope,
    NpmrcScopeConfigured,
    NpmrcGithubTokenValid,
    get_registry_host,
    npmrc_contains_scope_registry,
    npmrc_contains_auth_token,
    npmrc_scope_is_configured,
    get_npmrc_suggestion,
)

TEST_SCOPE_NAME = "my-org"
TEST_REGISTRY = "https://npm.pkg.github.com"
TEST_REGISTRY_HOST = "npm.pkg.github.com"
TEST_AUTH_TOKEN = "ghp_abc123xyz"


class TestNpmrcHelpers(unittest.TestCase):
    def test_get_registry_host(self):
        self.assertEqual(get_registry_host("https://npm.pkg.github.com"), "npm.pkg.github.com")
        self.assertEqual(get_registry_host("https://registry.npmjs.org"), "registry.npmjs.org")
        self.assertEqual(get_registry_host("https://npm.example.com/custom/path"), "npm.example.com")

    def test_npmrc_contains_scope_registry_matching(self):
        scope = NpmrcScope(name=TEST_SCOPE_NAME, registry=TEST_REGISTRY)
        lines = [
            f"@{TEST_SCOPE_NAME}:registry={TEST_REGISTRY}\n",
            f"//{TEST_REGISTRY_HOST}/:_authToken={TEST_AUTH_TOKEN}\n",
        ]
        self.assertTrue(npmrc_contains_scope_registry(lines, scope))

    def test_npmrc_contains_scope_registry_not_matching(self):
        scope = NpmrcScope(name=TEST_SCOPE_NAME, registry=TEST_REGISTRY)
        lines = [
            "@other-scope:registry=https://registry.npmjs.org\n",
        ]
        self.assertFalse(npmrc_contains_scope_registry(lines, scope))

    def test_npmrc_contains_scope_registry_empty_lines(self):
        scope = NpmrcScope(name=TEST_SCOPE_NAME, registry=TEST_REGISTRY)
        lines = []
        self.assertFalse(npmrc_contains_scope_registry(lines, scope))

    def test_npmrc_contains_auth_token_matching(self):
        lines = [
            f"@{TEST_SCOPE_NAME}:registry={TEST_REGISTRY}\n",
            f"//{TEST_REGISTRY_HOST}/:_authToken={TEST_AUTH_TOKEN}\n",
        ]
        self.assertTrue(npmrc_contains_auth_token(lines, TEST_REGISTRY))

    def test_npmrc_contains_auth_token_not_matching(self):
        lines = [
            f"@{TEST_SCOPE_NAME}:registry={TEST_REGISTRY}\n",
            "//registry.npmjs.org/:_authToken=some-token\n",
        ]
        self.assertFalse(npmrc_contains_auth_token(lines, TEST_REGISTRY))

    def test_npmrc_contains_auth_token_placeholder_rejected(self):
        lines = [
            f"//{TEST_REGISTRY_HOST}/:_authToken=UPDATE_WITH_TOKEN\n",
        ]
        self.assertFalse(npmrc_contains_auth_token(lines, TEST_REGISTRY))

    def test_npmrc_contains_auth_token_empty_rejected(self):
        lines = [
            f"//{TEST_REGISTRY_HOST}/:_authToken=\n",
        ]
        self.assertFalse(npmrc_contains_auth_token(lines, TEST_REGISTRY))

    def test_npmrc_scope_is_configured_without_auth(self):
        scope = NpmrcScope(name=TEST_SCOPE_NAME, registry=TEST_REGISTRY, require_auth_token=False)
        lines = [
            f"@{TEST_SCOPE_NAME}:registry={TEST_REGISTRY}\n",
        ]
        self.assertTrue(npmrc_scope_is_configured(lines, scope))

    def test_npmrc_scope_is_configured_with_auth_present(self):
        scope = NpmrcScope(name=TEST_SCOPE_NAME, registry=TEST_REGISTRY, require_auth_token=True)
        lines = [
            f"@{TEST_SCOPE_NAME}:registry={TEST_REGISTRY}\n",
            f"//{TEST_REGISTRY_HOST}/:_authToken={TEST_AUTH_TOKEN}\n",
        ]
        self.assertTrue(npmrc_scope_is_configured(lines, scope))

    def test_npmrc_scope_is_configured_with_auth_missing(self):
        scope = NpmrcScope(name=TEST_SCOPE_NAME, registry=TEST_REGISTRY, require_auth_token=True)
        lines = [
            f"@{TEST_SCOPE_NAME}:registry={TEST_REGISTRY}\n",
        ]
        self.assertFalse(npmrc_scope_is_configured(lines, scope))

    def test_npmrc_scope_is_configured_registry_missing(self):
        scope = NpmrcScope(name=TEST_SCOPE_NAME, registry=TEST_REGISTRY)
        lines = [
            f"//{TEST_REGISTRY_HOST}/:_authToken={TEST_AUTH_TOKEN}\n",
        ]
        self.assertFalse(npmrc_scope_is_configured(lines, scope))

    def test_get_npmrc_suggestion_without_auth(self):
        scope = NpmrcScope(name=TEST_SCOPE_NAME, registry=TEST_REGISTRY, require_auth_token=False)
        suggestion = get_npmrc_suggestion(scope)
        self.assertEqual(suggestion, f"@{TEST_SCOPE_NAME}:registry={TEST_REGISTRY}")

    def test_get_npmrc_suggestion_with_auth(self):
        scope = NpmrcScope(name=TEST_SCOPE_NAME, registry=TEST_REGISTRY, require_auth_token=True)
        suggestion = get_npmrc_suggestion(scope)
        expected = f"@{TEST_SCOPE_NAME}:registry={TEST_REGISTRY}\n//{TEST_REGISTRY_HOST}/:_authToken=UPDATE_WITH_TOKEN"
        self.assertEqual(suggestion, expected)


class TestNpmrcScopeConfigured(unittest.TestCase):
    @mock.patch("daktari.checks.npmrc.get_npmrc_contents")
    def test_returns_pass_when_scope_configured(self, mock_get_contents):
        mock_get_contents.return_value = [
            f"@{TEST_SCOPE_NAME}:registry={TEST_REGISTRY}\n",
            f"//{TEST_REGISTRY_HOST}/:_authToken={TEST_AUTH_TOKEN}\n",
        ]
        scope = NpmrcScope(name=TEST_SCOPE_NAME, registry=TEST_REGISTRY, require_auth_token=True)
        result = NpmrcScopeConfigured(scope).check()
        self.assertEqual(result.status, CheckStatus.PASS)

    @mock.patch("daktari.checks.npmrc.get_npmrc_contents")
    def test_returns_fail_when_scope_not_configured(self, mock_get_contents):
        mock_get_contents.return_value = [
            "@other-scope:registry=https://registry.npmjs.org\n",
        ]
        scope = NpmrcScope(name=TEST_SCOPE_NAME, registry=TEST_REGISTRY)
        result = NpmrcScopeConfigured(scope).check()
        self.assertEqual(result.status, CheckStatus.FAIL)

    @mock.patch("daktari.checks.npmrc.get_npmrc_contents")
    def test_returns_fail_when_file_not_found(self, mock_get_contents):
        mock_get_contents.side_effect = Exception("~/.npmrc does not exist")
        scope = NpmrcScope(name=TEST_SCOPE_NAME, registry=TEST_REGISTRY)
        result = NpmrcScopeConfigured(scope).check()
        self.assertEqual(result.status, CheckStatus.FAIL)
        self.assertIn("does not exist", result.summary)


class TestNpmrcGithubTokenValid(unittest.TestCase):
    @mock.patch("daktari.checks.npmrc.can_run_command")
    def test_returns_pass_when_token_valid(self, mock_can_run):
        mock_can_run.return_value = True
        result = NpmrcGithubTokenValid(scope_name=TEST_SCOPE_NAME, test_package="some-package").check()
        self.assertEqual(result.status, CheckStatus.PASS)
        mock_can_run.assert_called_once_with(f"npm view @{TEST_SCOPE_NAME}/some-package version")

    @mock.patch("daktari.checks.npmrc.can_run_command")
    def test_returns_fail_when_token_invalid(self, mock_can_run):
        mock_can_run.return_value = False
        result = NpmrcGithubTokenValid(scope_name=TEST_SCOPE_NAME, test_package="some-package").check()
        self.assertEqual(result.status, CheckStatus.FAIL)
        mock_can_run.assert_called_once_with(f"npm view @{TEST_SCOPE_NAME}/some-package version")


if __name__ == "__main__":
    unittest.main()



import unittest
from unittest import mock

from daktari.check import CheckStatus
from daktari.checks.pnpm import PnpmInstalled


class TestPnpm(unittest.TestCase):
    @mock.patch("daktari.check.can_run_command")
    def test_pnpm_installed_returns_pass_when_installed(self, mock_can_run_command):
        mock_can_run_command.return_value = True
        result = PnpmInstalled().check()
        self.assertEqual(result.status, CheckStatus.PASS)
        mock_can_run_command.assert_called_once_with("pnpm --version")

    @mock.patch("daktari.check.can_run_command")
    def test_pnpm_installed_returns_fail_when_not_installed(self, mock_can_run_command):
        mock_can_run_command.return_value = False
        result = PnpmInstalled().check()
        self.assertEqual(result.status, CheckStatus.FAIL)
        mock_can_run_command.assert_called_once_with("pnpm --version")


if __name__ == "__main__":
    unittest.main()



from daktari.check import Check, CheckResult
from daktari.os import OS

# We found that if /etc/hosts used a mixture of tabs and spaces
# it caused some applications to ignore /etc/hosts and perform DNS lookups instead.
# This check enforces a single space in any blankspace in /etc/hosts.


class EtcHostsFormattedCorrectly(Check):
    name = "etc.hosts.formattedCorrectly"
    description = "Check if /etc/hosts is formatted correctly with consistent use of tabs and spaces"

    def __init__(self):
        self.suggestions = {
            OS.OS_X: "Reformat /etc/hosts to use a single space consistently in blankspace:\n\n"
            + r"<cmd>sudo sed -Ei '' 's:( |\t)+: :g' /etc/hosts</cmd>",
            OS.UBUNTU: "Reformat /etc/hosts to use a single space consistently in blankspace:\n\n"
            + r"<cmd>sudo sed -Ei 's:( |\t)+: :g' /etc/hosts</cmd>",
        }

    def check(self) -> CheckResult:
        with open("/etc/hosts", "r") as f:
            content = f.read()
            if "\t" in content or "  " in content:
                return self.failed("/etc/hosts is uses tabs or multiple spaces in blankspace.")
            else:
                return self.passed("/etc/hosts is formatted correctly")



from typing import Optional


def get_tool_version_from_string(tool: str, file_content: str) -> Optional[str]:
    lines = file_content.split("\n")
    for line in lines:
        clean_line = line.split("#")[0].strip()
        if clean_line.startswith(tool + " "):
            return clean_line.split()[1]
    return None


def get_tool_version(tool: str) -> Optional[str]:
    with open(".tool-versions", "r") as f:
        file_content = f.read()
    return get_tool_version_from_string(tool, file_content)



import logging
import re
import os
from dataclasses import dataclass, replace, field
from pathlib import Path
from typing import Any, Dict, List, Optional

import yaml
from colors import red, yellow
from packaging import version
from yaml import YAMLError

from daktari import __version__
from daktari.check import Check
from daktari.check_utils import filter_out_checks_by_name
from daktari.resource_utils import get_resource
from daktari.result_printer import print_suggestion_text
from daktari.command_utils import CommandErrorException, run_command


@dataclass
class Config:
    min_version: Optional[str]
    title: Optional[str]
    checks: List[Check]
    ignored_checks: List[Check] = field(default_factory=list)
    quiet_mode: bool = False


version_regex = re.compile('daktari_version.*"([0-9.]+)"')
LOCAL_CONFIG_PATH = ".daktari-local.yaml"
LOCAL_CONFIG_TEMPLATE = "daktari-local-template.yaml"


def read_config(config_path: Path) -> Optional[Config]:
    raw_config = config_path.read_text()
    config = parse_raw_config(config_path, raw_config)
    if config is None:
        return config

    return apply_local_config(config)


def apply_local_config(config: Config) -> Optional[Config]:
    if not Path(LOCAL_CONFIG_PATH).is_file():
        return config

    try:
        with open(LOCAL_CONFIG_PATH, "rb") as local_config_file:
            local_config = yaml.safe_load(local_config_file)
    except YAMLError:
        print(red(f"❌  Failed to parse {LOCAL_CONFIG_PATH} - config is not valid YAML. Error follows."))
        logging.error(f"Exception reading {LOCAL_CONFIG_PATH}", exc_info=True)
        return None

    # E.g. the file has been entirely commented out
    if local_config is None:
        return config

    ignored_checks: List[str] = local_config.get("ignoredChecks", [])
    updated_config = remove_ignored_checks(config, ignored_checks)

    if local_config.get("alwaysQuiet", False):
        updated_config = replace(updated_config, quiet_mode=True)

    return updated_config


def write_local_config_template():
    contents = get_resource(LOCAL_CONFIG_TEMPLATE)
    with open(LOCAL_CONFIG_PATH, "w") as config_file:
        config_file.write(contents)

    print(
        f"A local config file has been generated at {LOCAL_CONFIG_PATH}. "
        f"Use this to override daktari behaviour - see the file for more details."
    )


def remove_ignored_checks(config: Config, ignored_check_names: List[str]) -> Config:
    remaining_checks = filter_out_checks_by_name(config.checks, set(ignored_check_names))
    remaining_check_names = {check.name for check in remaining_checks}
    ignored_checks = [check for check in config.checks if check.name not in remaining_check_names]
    return replace(config, checks=remaining_checks, ignored_checks=ignored_checks)


def parse_raw_config(config_path: Path, raw_config: str) -> Optional[Config]:
    if not check_version_compatibility(config_path, raw_config):
        return None

    variables: Dict[str, Any] = {}
    try:
        compiled_config = compile(raw_config, config_path, "exec")
        exec(compiled_config, variables)
    except Exception:
        print(red(f"❌  Failed to parse {config_path} - config is not valid."))
        logging.error(f"Exception reading {config_path}", exc_info=True)
        return None

    checks = variables.get("checks", [])
    title = variables.get("title", None)
    min_version = variables.get("daktari_version", None)
    return Config(min_version, title, checks)


def is_python_version_on_path() -> bool:
    path = (os.getenv("PATH") or "").lower()
    try:
        python3_version_raw = run_command(["python3", "--version"]).stdout.rstrip("\n")
        python3_version = re.search(r"3\.\d+\.\d+", python3_version_raw)
        if python3_version:
            python3_version_minor = ".".join(python3_version.group().split(".")[0:2])
        else:
            return False
    except CommandErrorException:
        logging.debug("Could not run 'python3 --version'", exc_info=True)
        return False

    if re.search("python.?3", path):
        if python3_version_minor in path:
            logging.debug(f"Python 3 version is {python3_version_minor}. This was detected as being on the PATH.")
            return True
        else:
            logging.debug(f"Python 3 version is {python3_version_minor}. This was NOT detected as being on the PATH.")
            logging.debug("It is likely the terminal will be unable to locate a newly installed version of Daktari")
            return False
    else:
        logging.debug("Python 3 does not appear to be on your PATH")
        logging.debug("Assuming Python is installed elsewhere in a non-versioned location")
        return True


def check_version_compatibility(config_path: Path, raw_config: str) -> bool:
    match = version_regex.search(raw_config)
    disclaimer = "Specifying daktari_version is recommended to ensure team members have compatible versions installed."
    if match is None:
        print(yellow(f"⚠️  No minimum version found in {config_path}. {disclaimer}"))
        return True

    my_version = version.parse(__version__)

    try:
        required_version = version.parse(match.group(1))
    except version.InvalidVersion:
        print(red(f"❌  Invalid daktari_version in {config_path}: {match.group(1)}"))
        return False

    if not isinstance(required_version, version.Version):
        print(red(f"❌  Invalid daktari_version in {config_path}: {required_version}"))
        return False

    if not is_python_version_on_path():
        print(
            yellow(
                "⚠️  A Python version was detected on your PATH. "
                "However the version on your PATH does not match the current version of Python 3 installed. "
                "It is possible Daktari is/will be installed under a directory not on your PATH. "
                "You may need to update your PATH to use the currently install Python version. "
                "E.g. /Users/username/Library/Python/3.11/bin"
            )
        )

    logging.debug(f"Doing version check. Mine [{my_version}], required [{required_version}]")
    if required_version > my_version:
        print(
            red(
                f"❌  Installed version of daktari [{my_version}] is "
                f"too old for this project (needs at least {required_version}). "
            )
        )
        print_suggestion_text("python3 -m pip install daktari -U")
        return False

    return True
